#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <cmath>
#include <signal.h>
#include <vector>
#include <iostream>
#include <iomanip>
#include <stdexcept>
#include <unistd.h> // usleep

#include "astrohaven.h"

const int MOTION_A = 0x80;
const int MOTION_B = 0x40;
const int TIME_OUT = 0x08;
const int BELTS_ERR= 0x04;
const int NO_220V  = 0x02;
const int OVERCURR = 0x01;
const int A_BOT    = 0x80;
const int A_TOP    = 0x40;
const int B_BOT    = 0x08;
const int B_TOP    = 0x04;

const int A_CLOSED = 0x20;
const int A_OPENED = 0x10;
const int B_CLOSED = 0x02;
const int B_OPENED = 0x01;

const int TIME_UNITS = 100;
const double CURR_UNIT = 3.45;

const std::string   subs_adome("Dome");
const std::string   sect_device("Devices");

const SComm Astrohaven::A_OPEN = 0x80;
const SComm Astrohaven::A_CLOSE =0x81;
const SComm Astrohaven::A_BACK = 0x8a;
const SComm Astrohaven::B_OPEN = 0x82;
const SComm Astrohaven::B_CLOSE =0x83;
const SComm Astrohaven::B_BACK = 0x8b;
const SComm Astrohaven::STOP   = 0x85;
const SComm Astrohaven::UNLOCK = 0x8c;
const SComm Astrohaven::LOCK   = 0x8e;

const SWord Astrohaven::SET_LIM_TIME =0x32;
const SWord Astrohaven::SET_ACTIVITY = 0x34;
const SWord Astrohaven::SET_UNLOCK_DEL = 0x36;
const SByte Astrohaven::SET_MAX_CURRENT = 0x28;
const SByte Astrohaven::MOVE_A_TO =   0x48;
const SByte Astrohaven::MOVE_B_TO =   0x4a;

const AByte Astrohaven::GET_TILT_A        = 0xe9;
const AByte Astrohaven::GET_TILT_B        = 0xeb;

const AWord Astrohaven::GET_POWER_VOLTAGE = 0xfc;
const AWord Astrohaven::GET_TEMPERATURE   = 0xfa;

pthread_t Astrohaven::Watch_thread = 0;

Astrohaven::Astrohaven( Config& cfg, RS485& r, const std::string& s ) : Module( cfg, r, s ) {
    sleep_time = cfg.inp<int>( sect_device, s, "ActivityCheck");
    lim_motion_time = cfg.inp<int>( sect_device, s, "LimitTime");
    open_release = cfg.inp<double>( sect_device, s, "OpenReleaseTime");
    close_release = cfg.inp<double>( sect_device, s, "CloseReleaseTime");
    max_current = cfg.inp<double>( sect_device, s, "CurrentLimit");
    A_tilt_open  = 91;
    A_tilt_close = 227;
    B_tilt_open  = 91;
    B_tilt_close = 227;
    try {
        sendWordCommand( SET_LIM_TIME, (int)(lim_motion_time*TIME_UNITS) );
        sendByteCommand( SET_MAX_CURRENT,  (int)(max_current*CURR_UNIT) );
        sendSimpleCommand( UNLOCK );
        lineWatchdog();
    }
    catch ( const std::exception& e ) {
        std::string init("Init: ");
        ErrlogSingl::instance().write( ERR_STD_EXCP, init + e.what() );
    }
//    std::cerr << "Constructor Astrohaven" << std::endl;
}

Astrohaven::~Astrohaven() {
    if( Watch_thread != 0 && pthread_cancel(Watch_thread) == ESRCH ) {
        ErrlogSingl::instance().write("Watch thread is terminated already");
        return;
    }
    if( Watch_thread != 0 && pthread_join(Watch_thread, NULL) != 0 ) {
        ErrlogSingl::instance().write(ERR_ALG, "Astrohaven Watch pthread_join failed");
    }
    try{
        close_b();
        usleep( 1000000 );
        close_a();
    }
    catch ( const std::exception& e ) {
        std::string init("Park: ");
        ErrlogSingl::instance().write( ERR_STD_EXCP, init + e.what() );
    }
//    std::cerr << "Destructor Astrohaven" << std::endl;
}

const std::string& Astrohaven::devStatus() {
    static const std::string undef = "UNDEFINED";
    static const std::string dclos = "CLOSED";
    static const std::string dopen = "OPENED";
    return ( A_CLOSED&getSensor() && B_CLOSED&sensor )? dclos :
           ( ( A_OPENED&sensor && B_OPENED&sensor )? dopen : undef );
}

double Astrohaven::sideA_position() {
    return (double)(A_tilt_close - tilt_A())/(A_tilt_close - A_tilt_open);
}

double Astrohaven::sideB_position() {
    return (double)(B_tilt_close - tilt_B())/(B_tilt_close - B_tilt_open);
}

int Astrohaven::time() {
    return (int)(2*lim_motion_time+2*open_release+2);
}

Astrohaven *Astrohaven::Handler::curts;

bool Astrohaven::Handler::expired = false;

void Astrohaven::Handler::catch_stop(int sig) {
    curts->sendSimpleCommand( STOP );
    ErrlogSingl::instance().write( "Process was externally stopped" );
    expired = true;
}

Astrohaven::Handler::Handler( Astrohaven* th) {
    expired = false;
    curts = th;
    signal( SIGUSR1, catch_stop );
}

void Astrohaven::close_a(){
    Handler h( this );
    nsleep( 0.5 );
    sendSimpleCommand( UNLOCK );
    sendWordCommand( SET_LIM_TIME, (int)(lim_motion_time*TIME_UNITS) );
    if( !(A_CLOSED&getSensor()) ) {
        sendSimpleCommand( A_CLOSE );
        while( MOTION_A&getStatus() ) nsleep( 1.0 );
        check_error( "side A closing" );
    }
    sendSimpleCommand( LOCK );
    A_tilt_close = tilt_A();
}

void Astrohaven::close_b(){
    Handler h( this );
    nsleep( 0.5 );
    sendSimpleCommand( UNLOCK );
    sendWordCommand( SET_LIM_TIME, (int)(lim_motion_time*TIME_UNITS) );
    if( !(B_CLOSED&getSensor()) ) {
        sendSimpleCommand( B_CLOSE );
        while( MOTION_B&getStatus() ) nsleep( 1.0 );
        check_error( "side B closing" );
    }
    sendSimpleCommand( LOCK );
    B_tilt_close = tilt_B();
}

void Astrohaven::open_a(){
    Handler h( this );
    nsleep( 0.5 );
    sendSimpleCommand( UNLOCK );
    if( !(A_OPENED&getSensor()) ) {
        sendWordCommand( SET_LIM_TIME, (int)(open_release*TIME_UNITS) );
        sendSimpleCommand( A_OPEN );
        nsleep( open_release+0.5 );
        if( A_CLOSED&getSensor() ) {
            ErrlogSingl::instance().write( "Side A doesn't move" );
        } else {
            sendWordCommand( SET_LIM_TIME, (int)(lim_motion_time*TIME_UNITS) );
            sendSimpleCommand( A_OPEN );
            while( MOTION_A&getStatus() )  nsleep( 1.0 );
            check_error( "side A opening" );
        }
    }
    sendSimpleCommand( LOCK );
    A_tilt_open  = tilt_A();
}

void Astrohaven::open_b(){
    Handler h( this );
    nsleep( 0.5 );
    sendSimpleCommand( UNLOCK );
    if( !(B_OPENED&getSensor()) ) {
        sendWordCommand( SET_LIM_TIME, (int)(open_release*TIME_UNITS) );
        sendSimpleCommand( B_OPEN );
        nsleep( open_release+0.5 );
        if( B_CLOSED&getSensor() ) {
            ErrlogSingl::instance().write( "Side B doesn't move" );
        } else {
            sendWordCommand( SET_LIM_TIME, (int)(lim_motion_time*TIME_UNITS) );
            sendSimpleCommand( B_OPEN );
            while( MOTION_B&getStatus() )  nsleep( 1.0 );
            check_error( "side B opening" );
        }
    }
    sendSimpleCommand( LOCK );
    B_tilt_open  = tilt_B();
}

// Dome completely opened: pos = 1;

void Astrohaven::move_a( double pos){
    Handler h( this );
    int starting = tilt_A();
    nsleep( 0.5 );
    sendSimpleCommand( UNLOCK );
    int target = A_tilt_close - pos*(A_tilt_close - A_tilt_open);
    sendWordCommand( SET_LIM_TIME, (int)(lim_motion_time*TIME_UNITS) );
    sendByteCommand( MOVE_A_TO, target );
    while( MOTION_A&getStatus() )  nsleep( 1.0 );
    sendSimpleCommand( LOCK );
    if( abs( tilt_A() - starting ) < 3 ) ErrlogSingl::instance().write("Side A doesn't move");
    if( abs( tilt_A() - target ) < 3 )   ErrlogSingl::instance().write("Side A reach the target");
}

void Astrohaven::move_b( double pos){
    Handler h( this );
    int starting = tilt_B();
    nsleep( 0.5 );
    sendSimpleCommand( UNLOCK );
    int target = B_tilt_close - pos*(B_tilt_close - B_tilt_open);
    sendWordCommand( SET_LIM_TIME, (int)(lim_motion_time*TIME_UNITS) );
    sendByteCommand( MOVE_B_TO, target );
    while( MOTION_B&getStatus() )  nsleep( 1.0 );
    sendSimpleCommand( LOCK );
    if( abs( tilt_B() - starting ) < 3 ) ErrlogSingl::instance().write("Side B doesn't move");
    if( abs( tilt_B() - target ) < 3 )   ErrlogSingl::instance().write("Side B reach the target");
}

int  Astrohaven::tilt_A() {
    return askByte( GET_TILT_A );
}

int  Astrohaven::tilt_B() {
    return askByte( GET_TILT_B );
}


bool Astrohaven::check_error( const std::string& s ) {
    getStatus();
    if( TIME_OUT&status )  ErrlogSingl::instance().write( "Time-out when " + s );
//    if( NO_220V&status )   ErrlogSingl::instance().write( "No 220 power: check fuse" );
//    if( OVERCURR&status )  ErrlogSingl::instance().write( "Motor was locked when " + s );
    return  TIME_OUT&status || NO_220V&status || OVERCURR&status;
}

bool Astrohaven::is_closed( bool A, bool B ){
    getSensor();
    if( (!(A_CLOSED&sensor) && A) || (!(B_CLOSED&sensor) && B) ) {
        ErrlogSingl::instance().wthrow( ERR_DEV_FUNC, "Dome isn't completely closed" );
        return( false );
    }
    ErrlogSingl::instance().write( "Dome is closed now" );
    return( true );
}

bool Astrohaven::is_opened( bool A, bool B ){
    getSensor();
    if( (!(A_OPENED&sensor) && A) || (!(B_OPENED&sensor) && B) ) {
        ErrlogSingl::instance().wthrow( ERR_DEV_FUNC, "Dome isn't completely opened" );
        return( false );
    }
    ErrlogSingl::instance().write( "Dome is opened now" );
    return( true );
}


void Astrohaven::lineWatchdog(){
    sendWordCommand( SET_ACTIVITY, sleep_time*TIME_UNITS );
    pthread_create( &Watch_thread, NULL, &Astrohaven::thread, (void*)this );
}

void *Astrohaven::thread( void* a ) {
    Astrohaven* ptr = (Astrohaven*)a;
    while(true) {
        try {
            if( ptr->connection() ) ptr->getSensor();       //    sleep_time/2
            ptr->nsleep(ptr->sleep_time/2);
        }
        catch ( RS485::ErrFatal& e ) {
            ErrlogSingl::instance().write( ERR_EXCHANGE, e.what() );
        }
        catch (const std::exception& e ) {
            ErrlogSingl::instance().write( ERR_STD_EXCP, e.what() );
        }
    }
    std::cerr << "Watch thread finished" << std::endl;
    return 0;
}

void Astrohaven::nsleep( double tval ) {
    timespec a;
    a.tv_sec  = (long)floor( tval );
    a.tv_nsec = (long)rint(1e+9*(tval-floor(tval)));
    if( nanosleep( &a, NULL ) < 0 ) ErrlogSingl::instance().write( "Action was interrupted" );
}

