#ifndef _EXCHANGE_H
#define _EXCHANGE_H
#include <stdexcept>
#include <sstream>
#include <sys/time.h>

const unsigned int IO_BUFF_SIZE = 64; // user/system buffer size

class RS485command {
protected:
    int cm;
public:
    RS485command( int code ) : cm(code) {}
    RS485command operator=( int code ){ return RS485command(code); }
    operator char() const {return (char)cm; }
};

class SComm : public RS485command {
public:
    SComm(int code) : RS485command(code) {}
};

class SByte : public RS485command {
public:
    SByte(int code) : RS485command(code) {}
};

class SWord : public RS485command {
public:
    SWord(int code) : RS485command(code) {}
};

class SData : public RS485command {
public:
    SData(int code) : RS485command(code) {}
};

class AByte : public RS485command {
public:
    AByte(int code) : RS485command(code) {}
};

class AWord : public RS485command {
public:
    AWord(int code) : RS485command(code) {}
};

class ALong : public RS485command {
public:
    ALong(int code) : RS485command(code) {}
};

class AData : public RS485command {
public:
    AData(int code) : RS485command(code) {}
};

class BaseRS485 {
protected:
    std::string estr;
private:
    static const int IVAL = -1;
    BaseRS485(const BaseRS485&);
    BaseRS485& operator=(const BaseRS485&);
public:
    enum com_stat {DONE, BUSY, NONE=0 };
    class ErrFatal : public std::runtime_error {
    public:
        ErrFatal( const std::string& mess ) : std::runtime_error(mess) {}
    };
    class ErrDriver : public std::runtime_error {
    public:
        ErrDriver( const std::string& mess ) : std::runtime_error(mess) {}
    };
    class ErrSignal : public std::runtime_error {
    public:
        ErrSignal( const std::string& mess ) : std::runtime_error(mess) {}
    };
public:
    BaseRS485() {} // std::cerr << "RS485 Base" << std::endl; }
    virtual ~BaseRS485() { } //std::cerr << "~RS485 Base" << std::endl;}
    virtual bool shadow() {return true;}
    virtual int askByte( unsigned int, const AByte) {return NONE;}
    virtual int askWord( unsigned int, const AWord) {return NONE;}
    virtual int askLong( unsigned int, const ALong) {return NONE;}
    virtual std::vector<unsigned char> askData( unsigned int, const AData) { return std::vector<unsigned char>(); }
    virtual com_stat sendSimpleCommand( unsigned int, const SComm) { return NONE; }
    virtual com_stat sendByteCommand( unsigned int, const SByte, char b) { return NONE; }
    virtual com_stat sendWordCommand( unsigned int, const SWord, short w) { return NONE; }
    virtual com_stat sendData( unsigned int, const SData, const std::vector<unsigned char>& ){ return NONE; }
    virtual unsigned int stat_packets(){ return 0; }
    virtual unsigned int stat_errors(){ return 0; }
    virtual std::string& protocol() { return estr; }
    virtual int readData( unsigned char*, unsigned char* ) { return 0; }
    virtual int readData( unsigned int&, unsigned int&, unsigned short* ) { return 0; }
    virtual void resetData() {};
};

class RS485impl : public BaseRS485 {
private:
    int id_mcua;
    unsigned char strcomm[IO_BUFF_SIZE];
    unsigned char strread[IO_BUFF_SIZE];
    unsigned char strdata[IO_BUFF_SIZE];
    unsigned int packet_number;
    unsigned int errors_number;
    unsigned int extranum;
    std::string device;
    int transaction(const unsigned char*, const int, unsigned char*);
    unsigned char low( short i) { return (unsigned char)(i&0x00ff); }
    unsigned char high( short i){ return (unsigned char)((i&0xff00)>>8); }
    std::string logstr;
    class RS485log {
        std::ostringstream s;
        RS485impl* ptr_up;
    public:
        RS485log( const unsigned char *inp,  const int nbyte, RS485impl* ptr );
        ~RS485log();
        void log( const std::string& t );
        void log( unsigned char *out, const int n );
    };
public:
    RS485impl( const std::string&, int r = 0 );
    ~RS485impl();
    bool shadow() {return false;}
    int askByte( unsigned int, const AByte );
    int askWord( unsigned int, const AWord );
    int askLong( unsigned int, const ALong );
    std::vector<unsigned char> askData( unsigned int, const AData );
    com_stat sendSimpleCommand( unsigned int, const SComm c);
    com_stat sendByteCommand( unsigned int, const SByte c, char b);
    com_stat sendWordCommand( unsigned int, const SWord c, short w);
    com_stat sendData( unsigned int, const SData c, const std::vector<unsigned char>& d );
    int sendVoid( const unsigned char* in, const int length, unsigned char* out);
    int readData( unsigned char*, unsigned char* );
    int readData( unsigned int& , unsigned int&, unsigned short* );
    unsigned int stat_packets(){ return packet_number; }
    unsigned int stat_errors(){ return errors_number; }
    std::string& protocol() { return logstr; }
    void resetData();
};

class RSTimer {
    struct itimerval old;
    struct itimerval end;
    static bool expired;
public:
    RSTimer( double );
    ~RSTimer();
    bool out() { return expired; }
private:
     static void catch_alarm(int sig);
};

#endif
