#ifndef __SINGLETON_H_
#define __SINGLETON_H_

/*
*   singleton.h is a part of carbon
*   Copyright (C) 2005  Matwey V. Kornilov
*
*   This program is free software; you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation; either version 2 of the License, or
*   (at your option) any later version.
*
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with this program; if not, write to the Free Software
*   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <cstdlib>
#include <stdexcept>

template< class T, template <class> class Traits > class Singleton{
public:
	typedef T BaseType;
	typedef Traits<T> TraitsType;
	
	inline static T& instance();
private:
	static void makeInstance();
	static void destroySingleton();
        
	Singleton();
        
	static T* s_instance;
	static bool s_destroyed;
};

template <class T> struct MayersSingletonTraits{
	static T* create(){
		static T obj;
		return &obj;
	}
	static void destroy( T* ){
	}
	template <class E> static void scheduleDestruction(T*, E destroyer){
		std::atexit(destroyer);
	}
	static void onDeadReference(){
		throw std::logic_error("Dead Reference Detected");
	}
};

template <class T> struct PhoenixSingletonTraits{
	static T* create(){
		static T obj;
		return &obj;
	}
	static void destroy( T* obj ){
		if(!s_destroyedOnce){
			obj->~T();
		}
		s_destroyedOnce=false;
	}
	template <class E> static void scheduleDestruction(T*, E destroyer){
		std::atexit(destroyer);
	}
	static void onDeadReference(){
		T* p=create();
		new(p) T;
	}
private:
	static bool s_destroyedOnce;
};

template< class T, template <class> class Traits > inline T& Singleton< T, Traits >::instance(){
	if(!s_instance){
		makeInstance();
	}
	return *s_instance;
}

template< class T, template <class> class Traits > void Singleton< T, Traits >::makeInstance(){
	if(!s_instance){
		if(s_destroyed){
			s_destroyed = false;
			Traits<T>::onDeadReference();
		}
		s_instance = Traits<T>::create();
		Traits<T>::scheduleDestruction(s_instance,&destroySingleton);
	}
}

template< class T, template <class> class Traits > void Singleton< T, Traits >::destroySingleton(){
	Traits<T>::destroy(s_instance);
	s_instance = 0;
	s_destroyed = true;
}
      
template <class T>                                 bool PhoenixSingletonTraits< T >::s_destroyedOnce = true;
template< class T, template <class> class Traits > T*   Singleton< T, Traits >::s_instance           = 0;
template< class T, template <class> class Traits > bool Singleton< T, Traits >::s_destroyed          = false;

#endif //  __SINGLETON_H_
